/*-
 * Copyright (c) 2015, 2021
 *      Bill Paul <wpaul@windriver.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <avr/io.h>

#include "dds.h"
#include "dds_in.h"

volatile static const char rcsid[] =
	"$Id: dds_in.c,v 1.33 2021/10/17 19:41:02 wpaul Exp $";

uint32_t freg;
uint8_t regsel;
uint8_t chipsel;

/*
 * This module handles the decoding of commands for the Kenwood F71022
 * DDS chips and recovers the frequency register values so that they can
 * be converted for use by the ADS9832 chips.
 *
 * According to the description in the Kenwood TS-950SDX service manual,
 * the F71022 chips use a clock frequency of 20.971520MHz and have two
 * 28-bit frequency registers. There is also a single 28-bit phase offset
 * register. Analysis using a logic probe revealed that the computer
 * on the digital unit generates values which are equal to the desired DDS
 * frequency multiplied by 12.8.
 *
 * The general formula for calculating the frequency register value for
 * DDS chips is as follows:
 *
 *                                     (register width)
 *             Freg = (Fout / Fclk) * 2
 *
 * If we substitute in the clock frequency of 20971520 for Fclk and use
 * 28 as the register width, then the calculation becomes:
 *
 *                                         28
 *             Freg = (Fout / 20971520) * 2
 *
 * What we need to do eventually is take the Freg value calculated by
 * the Kenwood computer for the F71022 chip and convert it into an Freg
 * value that produces the same frequency output from the ADS9832 chip.
 * The ADS9832 chip uses a similar scheme, except it has register width
 * of 32 bits instead of 28. So for that chip the calculation is:
 *
 *                                     32
 *             Freg = (Fout / Fclk) * 2
 *
 * Luckily, we're using the same Fclk for both chips, so that becomes:
 *
 *                                         32
 *             Freg = (Fout / 20971520) * 2
 *
 *
 * So basically, we want to do a calculation like this:
 *
 *                                              32
 * Freg(AD9832) = (Fout(Kenwood) / 20971520) * 2
 *
 * To get Fout(Kenwood), we reverse its Freg formula:
 *
 *                                               28
 * Fout(Kenwood) = (Freg(Kenwood) * 20971520) / 2
 *
 * If we substitue that in for Fout(Kenwood) in the first formula, we get:
 *
 *                                               28                 32
 * Freg(AD9832) = ((Freg(Kenwood) * 20971520) / 2  ) / 20971520) * 2
 *
 * The multiplication and division of 20971520 cancel out, so we're left
 * with:
 *
 *                                 28    32
 * Freg(AD9832) = Freg(Kenwood) / 2   * 2
 *
 *        32               28
 * Also, 2    divided by  2     is 16. So the final calculation is:
 *
 * Freq(ADS9832) = Freg(Kenwood) * 16
 *
 * Fortunately for us, this is a very simple calculation we can do with
 * integer math even with the limited capabilities of the ATmega328 processor.
 *
 * Also, analysis has also shown however that while the manual says that the
 * frequency registers are 28 bits wide, only 27 bits of the command word
 * are actually used for frequency data. Bits 0 through 26 were observed
 * to contain frequency data, while bit 27 (0x8000000) was used to select
 * between the two different frequency registers. Bit 28 (0x10000000) seems
 * to be used to select the phase shift register, however the computer only
 * programs this register once (at start-up) and always sets it to 0. This
 * indicates that the phase shift feature of the chip is not used. This
 * seems to limit the frequency range of the DDS chip to half the input
 * clock frequency. However this is not a problem in normal usage as in the
 * TS-950SDX the selected output range of the DDS chips is fairly low:
 *
 * IC3: 450KHz to 950KHz (500KHz range)
 * IC4: 790.76KHz to 1.29076MHz (500KHz range)
 *
 * Note that the actual desired range for IC4 is supposed to be 9.195MHz
 * to 9.695MHz. This is very close to the DDS chip's upper frequency range.
 * Perhaps for this reason, rather than programming IC4 to generate this
 * frequency directly, Kenwood chose to instead take the DDS output and
 * mix it with the "chopper" signal, which oscillates at 10.48576MHz. The
 * combination of 790.76KHz to 1.29076MHz and 10.48576MHz produces
 * 9.195MHz to 9.695MHz.
 */

/*
 * Note: we set the PORTB.0 pin to act as the clock output, so we
 * use pins 1, 2, 3 and 4 for input.
 */

#define DDS_CLK		0x02
#define DDS_DATA	0x04
#define DDS_ENB0	0x10
#define DDS_ENB1	0x08

void dds_in_init (void)
{
	return;
}

void dds_in_scan (void)
{
	uint8_t data;
	uint8_t clock_on = 0;
	uint8_t chip_on0 = 0;
	uint8_t chip_on1 = 0;
	uint8_t clock_change;
	uint8_t chip_change0;
	uint8_t chip_change1;
	uint32_t accum = 0;

again:

	/* Read the current data pin state */

	clock_change = FALSE;
	chip_change0 = FALSE;
	chip_change1 = FALSE;

	/* Sample once here to get clock and chip select state. */

	data = PINB;

	/* Check for clock signal assertion. */

	if (data & DDS_CLK) {
		if (clock_on == FALSE) {
			clock_change = TRUE;
			clock_on = TRUE;
		}
	} else  { 
		if (clock_on == TRUE) {
			clock_change = TRUE;
			clock_on = FALSE;
		}
	}

	/* Check for chip 0 ENB (latch) signal assertion. */

	if (data & DDS_ENB0) {
		if (chip_on0 == FALSE) {
			chip_change0 = TRUE;
			chip_on0 = TRUE;
		}
	} else {
		if (chip_on0 == TRUE) {
			chip_change0 = TRUE;
			chip_on0 = FALSE;
		}
	}

	/* Check for chip 0 ENB (latch) signal assertion. */

	if (data & DDS_ENB1) {
		if (chip_on1 == FALSE) {
			chip_change1 = TRUE;
			chip_on1 = TRUE;
		}
	} else {
		if (chip_on1 == TRUE) {
			chip_change1 = TRUE;
			chip_on1 = FALSE;
		}
	}

	/*
	 * Sample again here to get data state. This seems to
	 * help cope better with fast data setup times.
	 */

	data = PINB;

	/*
 	 * Accumulate a new data bit on a clock change. Note
	 * that the Kenwood uses an active low clock, so we
	 * test for when a state change to FALSE occurs to
	 * decide when to clock in a new data bit.
	 */

	if (clock_change == TRUE && clock_on == FALSE) {
		accum <<= 1;
		if (data & DDS_DATA)
			accum |= 1;
	}

	/*
	 * If the ENB0 or ENB1 bit has been set, that means the previous
	 * 30 bits we've accumulated are the new register setup command
	 * We calculate the frequency from the register data and stash it.
	 */

	if ((chip_change0 == TRUE && chip_on0 == TRUE) ||
	    (chip_change1 == TRUE && chip_on1 == TRUE)) {

		accum &= 0x1FFFFFFF;
		if (accum == 0)
			goto again;

		freg = (accum & DDS_CMD_FREQ);

		/*
		 * If the PREGSEL bit is set, then this commend is loading
		 * data into the phase offset register.
 		 * If the FREGSEL bit is set, then this command is
		 * programming the TX frequency register. If it's clear,
		 * then it's programming the RX register.
		 */

		accum &= DDS_CMD_PREGSEL|DDS_CMD_FREGSEL;

		switch (accum) {
			case DDS_CMD_PREGSEL:
				regsel = DDS_REG_PHASE;
				break;
			case DDS_CMD_FREGSEL:
				regsel = DDS_REG_TX;
				break;
			case 0:	
				regsel = DDS_REG_RX;
				break;
			default:
				/* Invalid condition */
				accum = 0;
				goto again;
				/* NOTREACHED */
				break;
		}

		if (chip_on0 == TRUE)
			chipsel = 0;

		if (chip_on1 == TRUE)
			chipsel = 1;

		/*
		 * freg now contains the frequency register value
		 * provided by the Kenwood computer. We'll use it
		 * to program the ADS9832 in the dds_out module.
                 */

		accum = 0;
		return;
	}

	goto again;

	/* NOTREACHED */
	return;
}
