/*-
 * Copyright (c) 2015
 *      Bill Paul <wpaul@windriver.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * This module implements driver support for the Analog Devices AD9832
 * direct digital synthesis chip. The datasheet for this chip is available
 * at:
 *
 * http://www.analog.com/media/en/technical-documentation/data-sheets/AD9832.pdf
 *
 * This code is also compatible with the AD9835 chip. The AD9832 is designed
 * to operate with a maximum clock frequency of 25MHz while the AD9835
 * supports clocks up to 50MHz. The Kenwood TS-950SDX uses a 20.971520MHz
 * clock, so the 50MHz chip is not required. (Though either one can be
 * used.)
 */

#include <stdint.h>
#include <avr/io.h>

#include "dds.h"
#include "dds_in.h"
#include "dds_out.h"

volatile static const char rcsid[] =
	"$Id: dds_out.c,v 1.29 2021/10/17 19:41:02 wpaul Exp $";

#define DDS_CLK		0x01
#define DDS_DATA	0x02
#define DDS_ENB		0x04

typedef struct _ovl {
	uint8_t		byte3;
	uint8_t		byte2;
	uint8_t		byte1;
	uint8_t		byte0;
} overlay;

static void dds_write0 (uint16_t);
static void dds_write1 (uint16_t);

static void dds_write0 (uint16_t cmd)
{
	uint16_t		i;

	PORTC &= (uint8_t)(~DDS_ENB);

	for (i = 0x8000; i; i >>= 1) {
		if (cmd & i)
			PORTC = DDS_DATA;
		PORTC |= DDS_CLK;
		PORTC = 0;
	}

	PORTC |= DDS_ENB;

	return;
}

static void dds_write1 (uint16_t cmd)
{
	uint16_t		i;

	PORTD &= (uint8_t)(~DDS_ENB);

	for (i = 0x8000; i; i >>= 1) {
		if (cmd & i)
			PORTD = DDS_DATA;
		PORTD |= DDS_CLK;
		PORTD = 0;
	}

	PORTD |= DDS_ENB;

	return;
}

void dds_out_init (void)
{
	PORTC = DDS_ENB;
	PORTD = DDS_ENB;

	/* Perform ADS9832 chip 0 initialziation. */

	/* Put chip into reset. */

	dds_write0 (AD_CTRL_INIT|AD_CTRL_INIT_RESET);

	/* Clear the phase registers */

	dds_write0 (AD_CMD_PHASE_DFRWR|AD_DST_PHASE0_BYTE0|0);
	dds_write0 (AD_CMD_PHASE_REGWR|AD_DST_PHASE0_BYTE1|0);
	dds_write0 (AD_CMD_PHASE_DFRWR|AD_DST_PHASE1_BYTE0|0);
	dds_write0 (AD_CMD_PHASE_REGWR|AD_DST_PHASE1_BYTE1|0);
	dds_write0 (AD_CMD_PHASE_DFRWR|AD_DST_PHASE2_BYTE0|0);
	dds_write0 (AD_CMD_PHASE_REGWR|AD_DST_PHASE2_BYTE1|0);
	dds_write0 (AD_CMD_PHASE_DFRWR|AD_DST_PHASE3_BYTE0|0);
	dds_write0 (AD_CMD_PHASE_REGWR|AD_DST_PHASE3_BYTE1|0);

	/* Clear the frequency registers */

	dds_write0 (AD_CMD_FREQ_DFRWR|AD_DST_FREQ0_BYTE0|0);
	dds_write0 (AD_CMD_FREQ_REGWR|AD_DST_FREQ0_BYTE1|0);
	dds_write0 (AD_CMD_FREQ_DFRWR|AD_DST_FREQ0_BYTE2|0);
	dds_write0 (AD_CMD_FREQ_REGWR|AD_DST_FREQ0_BYTE3|0);

	dds_write0 (AD_CMD_FREQ_DFRWR|AD_DST_FREQ1_BYTE0|0);
	dds_write0 (AD_CMD_FREQ_REGWR|AD_DST_FREQ1_BYTE1|0);
	dds_write0 (AD_CMD_FREQ_DFRWR|AD_DST_FREQ1_BYTE2|0);
	dds_write0 (AD_CMD_FREQ_REGWR|AD_DST_FREQ1_BYTE3|0);

	/* Set source selection to be done via external pins */

	dds_write0 (AD_CTRL_SELSRC);

	/* Take chip out of reset */

	dds_write0 (AD_CTRL_INIT);

	/* Perform ADS9832 chip 1 initialziation. */

	/* Put chip into reset. */

	dds_write1 (AD_CTRL_INIT|AD_CTRL_INIT_RESET);

	/* Clear the phase registers */

	dds_write1 (AD_CMD_PHASE_DFRWR|AD_DST_PHASE0_BYTE0|0);
	dds_write1 (AD_CMD_PHASE_REGWR|AD_DST_PHASE0_BYTE1|0);
	dds_write1 (AD_CMD_PHASE_DFRWR|AD_DST_PHASE1_BYTE0|0);
	dds_write1 (AD_CMD_PHASE_REGWR|AD_DST_PHASE1_BYTE1|0);
	dds_write1 (AD_CMD_PHASE_DFRWR|AD_DST_PHASE2_BYTE0|0);
	dds_write1 (AD_CMD_PHASE_REGWR|AD_DST_PHASE2_BYTE1|0);
	dds_write1 (AD_CMD_PHASE_DFRWR|AD_DST_PHASE3_BYTE0|0);
	dds_write1 (AD_CMD_PHASE_REGWR|AD_DST_PHASE3_BYTE1|0);

	/* Clear the frequency registers */

	dds_write1 (AD_CMD_FREQ_DFRWR|AD_DST_FREQ0_BYTE0|0);
	dds_write1 (AD_CMD_FREQ_REGWR|AD_DST_FREQ0_BYTE1|0);
	dds_write1 (AD_CMD_FREQ_DFRWR|AD_DST_FREQ0_BYTE2|0);
	dds_write1 (AD_CMD_FREQ_REGWR|AD_DST_FREQ0_BYTE3|0);

	dds_write1 (AD_CMD_FREQ_DFRWR|AD_DST_FREQ1_BYTE0|0);
	dds_write1 (AD_CMD_FREQ_REGWR|AD_DST_FREQ1_BYTE1|0);
	dds_write1 (AD_CMD_FREQ_DFRWR|AD_DST_FREQ1_BYTE2|0);
	dds_write1 (AD_CMD_FREQ_REGWR|AD_DST_FREQ1_BYTE3|0);

	/* Set source selection to be done via external pins */

	dds_write1 (AD_CTRL_SELSRC);

	/* Take chip out of reset */

	dds_write1 (AD_CTRL_INIT);

	return;
}

void dds_out_send (void)
{
	uint16_t accum[4];
	uint32_t reg;
	overlay * ovl;

	/*
	 * Ignore attempts to set the phase offset. The Kenwood
	 * radios always use a value of 0 as the phase offset.
	 */

	if (regsel == DDS_REG_PHASE)
		return;

	/*
	 * Since the AD9832 chip uses a 32-bit frequency register,
	 * the calculation to determine the register value for a given
 	 * output frequency is:
	 *
	 *                                   32
	 *       Freg = (Fout / 20971520) * 2
	 *
	 * Meanwhile, the calculation for the Kenwood F71022 chip is:
	 *
	 *                                   28
	 *       Freg = (Fout / 20971520) * 2
	 *
	 * It turns out we can easily determine the Freq value for the
	 * ADS9832 using the value from the F71022 using this simple
	 * calculation:
	 *
	 * Freg(ADS9832) = Freg(Kenwood) * 16
	 *
	 * (See the comments in dds_in.c for more details.)
	 */

	reg = freg * 16;

	if (regsel == DDS_REG_RX) {
		accum[0] = AD_CMD_FREQ_DFRWR|AD_DST_FREQ0_BYTE3;
		accum[1] = AD_CMD_FREQ_REGWR|AD_DST_FREQ0_BYTE2;
		accum[2] = AD_CMD_FREQ_DFRWR|AD_DST_FREQ0_BYTE1;
		accum[3] = AD_CMD_FREQ_REGWR|AD_DST_FREQ0_BYTE0;
	}

	if (regsel == DDS_REG_TX) {
		accum[0] = AD_CMD_FREQ_DFRWR|AD_DST_FREQ1_BYTE3;
		accum[1] = AD_CMD_FREQ_REGWR|AD_DST_FREQ1_BYTE2;
		accum[2] = AD_CMD_FREQ_DFRWR|AD_DST_FREQ1_BYTE1;
		accum[3] = AD_CMD_FREQ_REGWR|AD_DST_FREQ1_BYTE0;
	}

	ovl = (overlay *)&reg;

	accum[0] |= ovl->byte0;
	accum[1] |= ovl->byte1;
	accum[2] |= ovl->byte2;
	accum[3] |= ovl->byte3;

	if (chipsel == 0) {
		dds_write0 (accum[0]);
		dds_write0 (accum[1]);
		dds_write0 (accum[2]);
		dds_write0 (accum[3]);
	}

	if (chipsel == 1) {
		dds_write1 (accum[0]);
		dds_write1 (accum[1]);
		dds_write1 (accum[2]);
		dds_write1 (accum[3]);
	}

	return;
}
